// Iv2DViewer.c++

#include <stream.h>
#include "Iv2DViewer.h++"

static
void
error( char* fname, char* msg )
{
    cerr << "Iv2DViewer::" << fname << ": " << msg << endl;
}

void
Iv2DViewer::initialize( Widget wgt )
{
    if( inited )
       return;
    Widget parent = wgt;
    if( !wgt )
       parent = XtVaCreateManagedWidget( "Iv2DViewer",
					 vendorShellWidgetClass,
					 SoXt::getTopLevelWidget(),
					 NULL );
    if( !parent || NULL == parent )
       error( "initialize", "failed to create parent shell widget" );
    renderArea = new SoXtPlaneViewer( parent, "ToTo", TRUE, 
				      SoXtFullViewer::BUILD_NONE );
    renderArea -> setViewing( FALSE );
    renderArea -> show();
    if( !wgt )
       SoXt::show( parent );

    root -> ref();
    root -> addChild( camera );
    root -> addChild( keyEventNode );
    root -> addChild( selectRoot );
    selectRoot -> addChild( sceneRoot );

    root -> setName( "Iv2DViewer_root" );
    selectRoot -> setName( "Iv2DViewer_selectRoot" );
    sceneRoot -> setName( "Iv2DViewer_sceneRoot" );
    keyEventNode -> setName( "Iv2DViewer_keyEventNode" );

    selectRoot -> addSelectionCallback( selectCB, this );
    selectRoot -> addDeselectionCallback( deselectCB, this );
    keyEventNode -> addEventCallback( SoKeyboardEvent::getClassTypeId(),
				      keypressCB,
				      this );
    keyEventNode -> addEventCallback( SoMouseButtonEvent::getClassTypeId(),
				      mouseCB,
				      this );

    selectRoot -> policy = SoSelection::SINGLE;

    transDraggerRoot -> ref();
    transDraggerRoot -> addChild( transDraggerTransform );
    transDraggerRoot -> addChild( transDragger );

    scaleDraggerRoot -> ref();
    scaleDraggerRoot -> addChild( scaleDraggerTransform );
    scaleDraggerRoot -> addChild( scaleDragger );

    renderArea -> setSceneGraph(root);
    renderArea -> setTitle( "2-D Viewer" );
    camera -> height = 10.;
    camera -> viewportMapping = SoCamera::CROP_VIEWPORT_LINE_FRAME;
    camera -> aspectRatio = 1.;
    camera -> position.setValue( 0., 0., 10. );
    camera -> nearDistance.setValue( 1. );
    camera -> farDistance.setValue( 100. );
    // camera -> pointAt( SbVec3f( 0., 0., 0. ) );
    renderArea -> saveHomePosition();
    inited = 1;
}

Iv2DViewer::Iv2DViewer( Widget wgt, int build )
  : renderArea( 0 ),
    root( new SoSeparator ),
    camera( new SoOrthographicCamera ),
    selectRoot( new SoSelection ),
    sceneRoot( new SoSeparator ),
    keyEventNode( new SoEventCallback ),
    userKeyCB( 0 ),
    userKeyData( 0 ),
    userMouseCB( 0 ),
    userMouseData( 0 ),
    altkey( 0 ),
    objectCount( 0 ),
    selectCount( 0 ),
    objRootSelected( 0 ),
    inited( 0 ),
    // transDraggerRoot( new SoSeparator ),
    transDraggerRoot( new SoTransformSeparator ),
    transDraggerTransform( new SoTransform ),
    transDragger( new SoTranslate2Dragger ),
    userTransValueCB( 0 ),
    userTransValueData( 0 ),
    scaleDraggerRoot( new SoTransformSeparator ),
    scaleDraggerTransform( new SoTransform ),
    scaleDragger( new SoScale2Dragger ),
    userScaleValueCB( 0 ),
    userScaleValueData( 0 )
{
    if( build )
       initialize( wgt );
}

Iv2DViewer::~Iv2DViewer()
{
    root -> unref();
}

int
Iv2DViewer::addObject( SoNode* node )
{
    // Note: this is not the most efficient way, because we add a SEP
    // and SoTransform even to objects which may already have a suitable
    // internal SoTransform.
    // We plan to deal with only a few objects, so not a big concern
    // This scheme facilitates the easy selection/manipulation technique below

    SoSeparator* sep = new SoSeparator;
    SoTransform* xform = new SoTransform;
    sceneRoot -> addChild( sep );
    sep -> addChild( xform );
    sep -> addChild( node );
    sep -> setName( "Iv2DViewer_object" );
    xform -> setName( "Iv2DViewer_objectXform" );
    return( objectCount++ );
}

void
Iv2DViewer::removeObject( SoSeparator* objRoot )
{
    // meant to be called with the return of getSelectedObject() e.g.
    // i.e. has to be the object root separator
    SoPath* path = isObjectSelected( objRoot );
    if( path )
       selectRoot -> deselect( path );
    for( int i = 0; i < objectCount; i++ )
       if( objRoot == sceneRoot->getChild(i) )
	  {
	      sceneRoot -> removeChild(i);
	      objectCount--;
	      return;
	  }
}
 
SoTransform*
Iv2DViewer::getObjectTransform( int i )
{
    if( i < 0 || i >= objectCount )
       return 0;
    return( (SoTransform*)
	    ( ((SoSeparator*)(sceneRoot->getChild( i )))->getChild( 0 ) ) );
}

static
SoPath*
createPathToXform( SoPath* objectPath )
{
    // The first three elements of every object path are:
    // selectRoot -> sceneRoot -> objSep
    SoPath* pathToXform = objectPath -> copy( 0, 3 );
    pathToXform -> ref();
    // The SoTransform is the first child of objSep
    SoGroup* objRoot = ((SoGroup *)(pathToXform->getTail()));
    for( int i = 0; i < objRoot->getNumChildren(); i++ )
       if( objRoot->getChild(i)->isOfType( SoTransform::getClassTypeId() ) )
	  break;
    pathToXform->append( ((SoGroup *)(pathToXform->getTail()))->getChild(i) );
    pathToXform -> unrefNoDelete();
    return( pathToXform );
}

void
Iv2DViewer::selectCB( void *svptr, SoPath *selectionPath )
{
    Iv2DViewer* sv = (Iv2DViewer*) svptr;
    SoPath* xformPath = createPathToXform( selectionPath );
    xformPath -> ref();
    SoTransform* objXform = (SoTransform*)xformPath->getTail();
    sv->objRootSelected = (SoSeparator*) xformPath->getNode(2);
    SbVec3f objLoc = objXform->translation.getValue();
    if( !sv->altkey )
       {
	   sv->transDragger -> translation.setValue
	      ( objLoc[0], objLoc[1], 1. );
	   sv->transDragger -> 
	      addValueChangedCallback( transDraggerValueCB, sv );
	   sv->objRootSelected -> insertChild( sv->transDraggerRoot, 0 );
       }
    else
       {
	   SbVec3f objScale = objXform->scaleFactor.getValue();
	   sv->scaleDragger -> scaleFactor.setValue
	      ( objScale[0], objScale[1], 1. );
	   sv->scaleDraggerTransform->translation.setValue
	      ( objLoc[0], objLoc[1], 1. );
	   sv->scaleDragger -> 
	      addValueChangedCallback( scaleDraggerValueCB, sv );
	   sv->objRootSelected -> insertChild( sv->scaleDraggerRoot, 0 );
       }
    sv -> selectCount++;
    xformPath -> unref();
}

void
Iv2DViewer::deselectCB( void *svptr, SoPath *deselectionPath )
{
    Iv2DViewer* sv = (Iv2DViewer*) svptr;
    SoPath* xformPath = createPathToXform( deselectionPath );
    xformPath -> ref();
    sv->objRootSelected = (SoSeparator*) xformPath->getNode(2);
    SoNode* dragNode = sv->objRootSelected -> getChild(0);
    if( dragNode == sv->transDraggerRoot )
       {
	   sv->objRootSelected -> removeChild( sv->transDraggerRoot );
	   sv->transDragger ->
	      removeValueChangedCallback( transDraggerValueCB, sv);
       }
    else if ( dragNode == sv->scaleDraggerRoot )
       {
	   sv->objRootSelected -> removeChild( sv->scaleDraggerRoot );
	   sv->scaleDragger ->
	      removeValueChangedCallback( scaleDraggerValueCB, sv );
       }
    sv -> selectCount--;
    if( !sv->selectCount )
       sv -> objRootSelected = 0;
    if( sv->selectCount < 0 )
       error( "deselectCB", "selectCount < 0" );
    xformPath -> unref();
}

void
Iv2DViewer::keypressCB( void* svptr, SoEventCallback* node )
{
    Iv2DViewer* sv = (Iv2DViewer*) svptr;
    if( sv->userKeyCB )
       if( sv->userKeyCB( sv->userKeyData, node ) )
	  return;
    const SoEvent *event = node -> getEvent();
    if( SO_KEY_PRESS_EVENT( event, Q ) )
       exit(0);
    if( SO_KEY_PRESS_EVENT( event, LEFT_SHIFT ) ||
	SO_KEY_PRESS_EVENT( event, RIGHT_SHIFT ) )
       sv -> altkey = 1;
    if( SO_KEY_RELEASE_EVENT( event, LEFT_SHIFT ) ||
	SO_KEY_RELEASE_EVENT( event, RIGHT_SHIFT ) )
       sv -> altkey = 0;
}

void
Iv2DViewer::mouseCB( void* svptr, SoEventCallback* node )
{
    Iv2DViewer* sv = (Iv2DViewer*) svptr;
    if( sv->userMouseCB )
       if( sv->userMouseCB( sv->userMouseData, node ) )
	  return;
#if 0
    const SoEvent *event = node -> getEvent();
    if( SO_MOUSE_PRESS_EVENT( event, BUTTON1 ) )
       cerr << "Button 1" << endl;
#endif
}

int
Iv2DViewer::printPS( FILE* fp, int whitebg )
{
    SbViewportRegion viewport( renderArea->getViewportRegion() );
    viewport.setPixelsPerInch( 144. );
    SoOffscreenRenderer osr( viewport );
    if( whitebg )
       osr.setBackgroundColor( SbColor( 1., 1., 1. ) );
    osr.render( root );
    osr.writeToPostScript( fp );
    return 0;
}

#if 0
static
ostream&
printVec( ostream& os, const SbVec3f& v )
{
    return( os << v[0] << ", " << v[1] << ", " << v[2] << endl );
}
#endif

void
Iv2DViewer::transDraggerValueCB( void* vptr, SoDragger* dragger )
{
    Iv2DViewer* tdv = (Iv2DViewer*) vptr;
    // objRoot, the last two children are: Xform -- Obj
    int n = tdv->objRootSelected -> getNumChildren();
    SoTranslate2Dragger* std = (SoTranslate2Dragger*) dragger;
    Iv2DViewerCBData cbdata;
    cbdata.value = std->translation.getValue();
    cbdata.object = tdv->objRootSelected->getChild(n-1);
    if( tdv->userTransValueCB )
       if( tdv->userTransValueCB( tdv->userTransValueData, &cbdata ) )
	  return;
    SoTransform* xform = (SoTransform*) tdv->objRootSelected->getChild(n-2);
    xform->translation.setValue( cbdata.value[0], cbdata.value[1], 0. );
}

void
Iv2DViewer::scaleDraggerValueCB( void* vptr, SoDragger* dragger )
{
    Iv2DViewer* tdv = (Iv2DViewer*) vptr;
    // SoTransform* xform = (SoTransform*) vptr;
    int n = tdv->objRootSelected -> getNumChildren();
    SoScale2Dragger* std = (SoScale2Dragger*) dragger;
    Iv2DViewerCBData cbdata;
    cbdata.value = std->scaleFactor.getValue();
    cbdata.object = tdv->objRootSelected->getChild(n-1);
    if( tdv->userScaleValueCB )
       if( tdv->userScaleValueCB( tdv->userScaleValueData, &cbdata ) )
	  return;
    SoTransform* xform = (SoTransform*) tdv->objRootSelected->getChild(n-2);
    xform->scaleFactor.setValue( cbdata.value[0], cbdata.value[1], 1. );
}

SoPath*
Iv2DViewer::isObjectSelected( SoSeparator* objRoot )
{
    SoPath* path;
    int nsel = selectRoot -> getNumSelected();
    for( int i = 0; i < nsel; i++ )
       {
	   path = selectRoot->getPath(i);
	   SoSeparator* pathObjRoot = (SoSeparator*) path->getNode(2);
	   if( pathObjRoot == objRoot )
	      return path;
       }
    return 0;
}
