// Iv3DViewer.c++

#include <stream.h>
#include "Iv3DViewer.h++"

static
void
error( char* fname, char* msg )
{
    cerr << "Iv3DViewer::" << fname << ": " << msg << endl;
}

void
Iv3DViewer::initialize( Widget wgt )
{
    if( inited )
       return;
    Widget parent = wgt;
    if( !wgt )
       parent = XtVaCreateManagedWidget( "Iv3DViewer",
					 vendorShellWidgetClass,
					 SoXt::getTopLevelWidget(),
					 NULL );
    if( !parent || NULL == parent )
       error( "initialize", "failed to create parent shell widget" );
    examiner = new SoXtExaminerViewer( parent );
    examiner -> show();
    if( !wgt )
       SoXt::show( parent );

    root -> ref();
    root -> addChild( view3Dcamera );
    root -> addChild( keyEventNode );
    root -> addChild( selectRoot );
    selectRoot -> addChild( sceneRoot );

    root -> setName( "Iv3DViewer_root" );
    selectRoot -> setName( "Iv3DViewer_selectRoot" );
    sceneRoot -> setName( "Iv3DViewer_sceneRoot" );
    keyEventNode -> setName( "Iv3DViewer_keyEventNode" );

    selectRoot -> addSelectionCallback( selectCB, this );
    selectRoot -> addDeselectionCallback( deselectCB, this );
    keyEventNode -> addEventCallback( SoKeyboardEvent::getClassTypeId(),
				      keypressCB,
				      this );
    examiner -> setSceneGraph(root);
    examiner -> setTitle( "3-D Viewer" );
    view3Dcamera -> position.setValue( 0., 0., 10. );
    view3Dcamera -> pointAt( SbVec3f( 0., 0., 0. ) );
    examiner -> saveHomePosition();

    overlay -> attachTo( examiner );
    inited = 1;
}

Iv3DViewer::Iv3DViewer( Widget wgt, int build )
  : examiner( 0 ),
    root( new SoSeparator ),
    view3Dcamera( new SoPerspectiveCamera ),
    selectRoot( new SoSelection ),
    sceneRoot( new SoSeparator ),
    keyEventNode( new SoEventCallback ),
    userKeyCB( 0 ),
    userKeyData( 0 ),
    altkey( 0 ),
    objectCount( 0 ),
    selectCount( 0 ),
    objRootLastSelected( 0 ),
    inited( 0 ),
    overlay( new IvCameraOverlay )
  // titleRoot( 0 ),
  // titleXform( 0 ),
  // titleText( 0 )
{
    if( build )
       initialize( wgt );
}

Iv3DViewer::~Iv3DViewer()
{
    root -> unref();
}

int
Iv3DViewer::addObject( SoNode* node )
{
    // Note: this is not the most efficient way, because we add a SEP
    // and SoTransform even to objects which may already have a suitable
    // internal SoTransform.
    // We plan to deal with only a few objects, so not a big concern
    // This scheme facilitates the easy selection/manipulation technique below

    SoSeparator* sep = new SoSeparator;
    SoTransform* xform = new SoTransform;
    sceneRoot -> addChild( sep );
    sep -> addChild( xform );
    sep -> addChild( node );
    sep -> setName( "Iv3DViewer_object" );
    xform -> setName( "Iv3DViewer_objectXform" );
    return( objectCount++ );
}

void
Iv3DViewer::removeObject( SoSeparator* objRoot )
{
    // meant to be called with the return of getSelectedObject() e.g.
    // i.e. has to be the object root separator
    for( int i = 0; i < objectCount; i++ )
       if( objRoot == sceneRoot->getChild(i) )
	  {
	      sceneRoot -> removeChild(i);
	      objectCount--;
	      return;
	  }
}
 
void
Iv3DViewer::removeObject( int i )
{
    // meant to be called with the return of getSelectedObject() e.g.
    // i.e. has to be the object root separator
    sceneRoot -> removeChild( i );
    objectCount--;
}
 
SoTransform*
Iv3DViewer::getObjectTransform( int i )
{
    if( i < 0 || i >= objectCount )
       return 0;
    return( (SoTransform*)
	    ( ((SoSeparator*)(sceneRoot->getChild( i )))->getChild( 0 ) ) );
}

SoSeparator*
Iv3DViewer::getObject( int i )
{
    if( i < 0 || i >= objectCount )
       return 0;
    return( (SoSeparator*)(sceneRoot->getChild( i )) );
}

static
SoPath*
createPathToXform( SoPath* objectPath )
{
    // The first three elements of every object path are:
    // selectRoot -> sceneRoot -> objSep
    SoPath* pathToXform = objectPath -> copy( 0, 3 );
    pathToXform -> ref();
    // The SoTransform is the first child of objSep
    pathToXform->append( ((SoGroup *)(pathToXform->getTail()))->getChild(0) );
    pathToXform -> unrefNoDelete();
    return( pathToXform );
}

void
Iv3DViewer::selectCB( void *svptr, SoPath *selectionPath )
{
    // cerr << "selectCB:" << endl;
    Iv3DViewer* sv = (Iv3DViewer*) svptr;
    SoPath* xformPath = createPathToXform( selectionPath );
    xformPath -> ref();
    SoTransformManip* manipulator = ( sv->altkey )
       ? (SoTransformManip*) new SoHandleBoxManip
       : (SoTransformManip*) new SoTransformBoxManip;
    manipulator -> setName( "my manipulator" );
    manipulator -> replaceNode( xformPath );
    sv -> selectCount++;
    sv -> objRootLastSelected = (SoSeparator*) xformPath->getNode(2);
    xformPath -> unref();
}

void
Iv3DViewer::deselectCB( void *svptr, SoPath *deselectionPath )
{
    // cerr << "deselectCB:" << endl;
    Iv3DViewer* sv = (Iv3DViewer*) svptr;
    SoPath* xformPath = createPathToXform( deselectionPath );
    xformPath -> ref();
    ((SoTransformManip *)(xformPath->getTail()))
       -> replaceManip( xformPath, new SoTransform );
    sv -> selectCount--;
    if( !sv->selectCount )
       sv -> objRootLastSelected = 0;
    if( sv->selectCount < 0 )
       error( "deselectCB", "selectCount < 0" );
    xformPath -> unref();
}

void
Iv3DViewer::keypressCB( void* svptr, SoEventCallback* node )
{
    Iv3DViewer* sv = (Iv3DViewer*) svptr;
    if( sv->userKeyCB )
       if( sv->userKeyCB( sv->userKeyData, node ) )
	  return;
    const SoEvent *event = node -> getEvent();
    if( SO_KEY_PRESS_EVENT( event, Q ) )
       exit(0);
    if( SO_KEY_PRESS_EVENT( event, V ) )
       sv -> printCameraTransform( cerr );
    if( SO_KEY_PRESS_EVENT( event, LEFT_SHIFT ) ||
	SO_KEY_PRESS_EVENT( event, RIGHT_SHIFT ) )
       sv -> altkey = 1;
    if( SO_KEY_RELEASE_EVENT( event, LEFT_SHIFT ) ||
	SO_KEY_RELEASE_EVENT( event, RIGHT_SHIFT ) )
       sv -> altkey = 0;
}

#if 0
static
ostream&
printVec( ostream& os, const SbVec2s& v )
{
    return( os << v[0] << ", " << v[1] );
}

static
ostream&
printVec( ostream& os, const SbVec2f& v )
{
    return( os << v[0] << ", " << v[1] );
}
#endif

int
Iv3DViewer::printPS( FILE* fp )
{
    SbViewportRegion viewport( examiner->getViewportRegion() );
    viewport.setPixelsPerInch( 144. );
    SoOffscreenRenderer osr( viewport );
    osr.setBackgroundColor( SbColor( 1., 1., 1. ) );
    osr.render( root );
    osr.writeToPostScript( fp );
    return 0;
}

int
Iv3DViewer::writeRGB( FILE* fp, int w, int h )
{
    SbViewportRegion viewport( examiner->getViewportRegion() );
    if( w )
       viewport.setWindowSize( w, h );
    // viewport.setPixelsPerInch( 144. );
    SoOffscreenRenderer osr( viewport );
    // osr.setBackgroundColor( SbColor( 1., 1., 1. ) );
    osr.render( root );
    osr.writeToRGB( fp );
    return 0;
}

ostream&
Iv3DViewer::printCameraTransform( ostream& os )
{
    float tmp[4];
    view3Dcamera -> position.getValue().getValue( tmp[0], tmp[1], tmp[2] );
    os << tmp[0] << "  " <<  tmp[1] << "  " <<  tmp[2] << endl;
    view3Dcamera -> orientation.getValue().getValue( tmp[0], tmp[1], 
						     tmp[2], tmp[3] );
    os << tmp[0] << " " <<  tmp[1] << " " <<  tmp[2] << " " << tmp[3] << endl;
    os << view3Dcamera->heightAngle.getValue() << endl;
    return os;
}

istream&
Iv3DViewer::scanCameraTransform( istream& is )
{
    float tmp[4];
    is >> tmp[0] >> tmp[1] >> tmp[2];
    view3Dcamera -> position.setValue( tmp[0], tmp[1], tmp[2] );
    is >> tmp[0] >> tmp[1] >> tmp[2] >> tmp[3];
    view3Dcamera -> orientation.setValue( tmp[0], tmp[1], tmp[2], tmp[3] );
    is >> tmp[0];
    view3Dcamera -> heightAngle.setValue( tmp[0] );
    return is;
}

#if 0
void
Iv3DViewer::setTitle( const char* s, int x, int y, int z )
{
    if( !titleRoot )
       {
	   titleRoot = new SoSeparator;
	   titleXform = new SoTransform;
	   titleText = new SoText2;
	   root -> addChild( titleRoot );
	   titleRoot -> addChild( titleXform );
	   titleRoot -> addChild( titleText );
       }
    titleText -> string.setValue( s );
    titleXform -> translation.setValue( x, y, z );
}
#endif
