// IvCameraView.c++

#include <stream.h>
#include "IvCameraView.h++"
#include <Inventor/Xt/SoXt.h>

const int IvCameraView::defaultHeight = 240;
const int IvCameraView::defaultWidth = 320;

IvCameraView::IvCameraView( Widget wgt )
  : height( defaultHeight ),
    width( defaultWidth ),
    renderArea( new SoXtRenderArea(wgt) ),
    root( new SoSeparator ),
    camera( new SoPerspectiveCamera ),
    light( new SoSpotLight ),
    lightModel( new SoLightModel ),
    sceneRoot( new SoTransformSeparator ),
    coordinateTransform( new SoTransform ),
    keyEventNode( new SoEventCallback ),
    userKeyCB( 0 ),
    userKeyData( 0 ),
    userMouseCB( 0 ),
    userMouseData( 0 ),
    backgroundData( 0 ),
    backgroundNode( new SoCallback ),
    interiorOrientationSensor( 0 ),
    relativeOrientationSensor( 0 )
{
    if( !renderArea || !root || !camera || !light || !lightModel ||
	!sceneRoot || !coordinateTransform )
       fprintf( stderr, "IvCameraView: init failed\n" ), exit(1);

    root -> ref();
    root -> addChild( camera );
    root -> addChild( backgroundNode );
    root -> addChild( light );
    root -> addChild( lightModel );
    root -> addChild( sceneRoot );
    root -> addChild( keyEventNode );
    sceneRoot -> addChild( coordinateTransform );
    root -> setName( "IvCameraView_root" );
    camera -> setName( "IvCameraView_camera" );
    light -> setName( "IvCameraView_light" );
    lightModel -> setName( "IvCameraView_lightModel" );
    sceneRoot -> setName( "IvCameraView_sceneRoot" );
    coordinateTransform -> setName( "IvCameraView_coordinateTransform" );
    keyEventNode -> setName( "IvCameraView_keyEventNode" );
    keyEventNode -> addEventCallback( SoKeyboardEvent::getClassTypeId(),
				      keypressCB,
				      this );
    keyEventNode -> addEventCallback( SoMouseButtonEvent::getClassTypeId(),
				      mousepressCB,
				      this );
    light -> location.setValue( 0., 0., -1 );
    light -> direction.setValue( 0., 0., 1 );
    renderArea -> setSceneGraph( root );
    renderArea -> setTitle( "Camera" );
    renderArea -> show();
}

IvCameraView::~IvCameraView()
{
    root -> unref();
}

void
IvCameraView::setBackground( void* data, GLenum format )
{
    if( !data )
       return;
    if( !backgroundData )
       backgroundNode -> setCallback( backgroundCB, this );
    backgroundData = data;
    backgroundFormat = format;
}

void
IvCameraView::setInteriorOrientation( float beta )
{
    camera -> viewportMapping = SoCamera::CROP_VIEWPORT_FILL_FRAME;
    camera -> position.setValue( 0., 0., -1./beta );
    camera -> orientation.setValue( 1., 0., 0., 0. );
    // these numbers are good for Video
    // camera -> heightAngle = 2. * atan( .375 * beta );
    // camera -> aspectRatio = 4./3.;
    float halfHeight = float(height) / float(width) / 2.;
    camera -> heightAngle = 2. * atan( halfHeight * beta );
    camera -> aspectRatio = float(width) / float(height);
    camera -> nearDistance = .1/ beta;
    camera -> farDistance = 100.;
}

void
IvCameraView::setCoordinateTransform( float tx, float ty, float tz, 
				      float q0, float q1, float q2, float q3 )
{
    SbVec3f scale( 1., 1., 1. );
    SbMatrix lrxform;
    SbRotation rot( q1, q2, q3, q0 );
    SbVec3f tr( tx, ty, tz );
    lrxform.setTransform( tr, rot, scale );
    coordinateTransform -> setMatrix( lrxform );
}
    
void
IvCameraView::attachInteriorOrientation( SoPerspectiveCamera* pc )
{
    if( interiorOrientationSensor )
       {
	   interiorOrientationSensor -> detach();
	   delete interiorOrientationSensor;
       }
    interiorOrientationSensor = new SoNodeSensor
       ( &interiorOrientationSensorCB, this );
    interiorOrientationSensor -> attach( pc );
}

void
IvCameraView::attachReferenceTransform( SoTransform* rt )
{
    if( relativeOrientationSensor )
       {
	   relativeOrientationSensor -> detach();
	   delete relativeOrientationSensor;
       }
    relativeOrientationSensor = new SoNodeSensor
       ( &referenceTransformSensorCB, this );
    relativeOrientationSensor -> attach( rt );
}

void
IvCameraView::attachCoordinateTransform( SoTransform* ct )
{
    if( relativeOrientationSensor )
       {
	   relativeOrientationSensor -> detach();
	   delete relativeOrientationSensor;
       }
    relativeOrientationSensor = new SoNodeSensor
       ( coordinateTransformSensorCB, this );
    relativeOrientationSensor -> attach( ct );
}

static
inline
int
colorEq( unsigned char* c1, unsigned char* c2, int tol )
{
    int diff;
    for( int i = 0; i < 3; i++ )
       {
	   diff = c1[i] - c2[i];
	   if( diff < -tol || diff > tol )
	      return 0;
       }
    return 1;
}

int
IvCameraView::findColorBlob( SbColor& color, float* mean, float* cov )
{
    // (1) Get the buffer

    SoOffscreenRenderer myRenderer( SbViewportRegion( width, height ) );
    lightModel -> model.setValue( SoLightModel::BASE_COLOR );
    myRenderer.render( root );
    lightModel -> model.setValue( SoLightModel::PHONG );
    unsigned char* buffer = myRenderer.getBuffer();
    
    // (2) Find the pixels having the right color and calculate the sums

    unsigned char* ucp = buffer;
    unsigned char ucolor[3];
    ucolor[0] = (unsigned char) limit( color[0] * 255., 0., 255. );
    ucolor[1] = (unsigned char) limit( color[1] * 255., 0., 255. );
    ucolor[2] = (unsigned char) limit( color[2] * 255., 0., 255. );
    double xsum = 0.;
    double x2sum = 0.;
    double ysum = 0.;
    double y2sum = 0.;
    double xysum = 0.;
    long int count = 0;
    for( int y = 0; y < height; y+=2 )
       {
	   for( int x = 0; x < width; x+=2 )
	      {
		  if( colorEq( ucp, ucolor, 2 ) )
		     {
			 xsum += x;
			 x2sum += x*x;
			 ysum += y;
			 y2sum += y*y;
			 xysum += x*y;
			 count++;
		     }
		  ucp += 6;
	      }
	   ucp += width*3;
       }
    if( !count )
       return 1;

    // (4) translate from stupid graphics pixel coords to more intelligent
    //     and useful vision coords
    double xmid = width / 2.;
    double ymid = height / 2.;
    double a = 1. / width;
    // These are the first and second moments, in pixels
    xsum /= count;
    ysum /= count;
    x2sum /= count;
    y2sum /= count;
    xysum /= count;
    // These are the first and second moments, after linear xform
    mean[0] = a * ( xsum - xmid );
    mean[1] = -a * ( ysum - ymid );
    double Ex2 = a*a*( x2sum - 2*xsum*xmid + xmid*xmid );
    double Exy = a*a*( -xysum + xsum*ymid + xmid*ysum - xmid*ymid );
    double Ey2 = a*a*( y2sum - 2*ysum*ymid + ymid*ymid );
    cov[0] = Ex2 - mean[0]*mean[0];
    cov[1] = cov[2] = Exy - mean[0]*mean[1];
    cov[3] = Ey2 - mean[1]*mean[1];
    return 0;
}

#if 0
void
IvCameraView::updateBackground()
{
    GLfloat current[4] = { -.5, -.375, 0., 1. };
	   glPushAttrib( GL_ALL_ATTRIB_BITS );
	   glDisable( GL_LIGHTING );  // so we don't have to set normals
	   glDepthMask( GL_FALSE );
	   glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
	   glPixelStorei( GL_UNPACK_ROW_LENGTH, width );
	   glRasterPos4fv( current );
	   glPixelZoom( 1, -1 );
	   glDrawPixels( width,
			 height, 
			 // GL_RGB,
			 GL_ABGR_EXT,
			 GL_UNSIGNED_BYTE,
			 backgroundData );
	   glDepthMask( GL_TRUE );
	   glEnable( GL_LIGHTING );   
	   glPopAttrib();
}
#endif

void
IvCameraView::backgroundCB( void* icvptr, SoAction* action )
{
    IvCameraView* icv = (IvCameraView*) icvptr;
    // if( action->isOfType( SoGLRenderAction::getClassTypeId() ) )
    // icv -> updateBackground();
    // GLfloat current[4] = { -.5, -.375, 0., 1. };
    float halfHeight = float(icv->height) / float(icv->width) / 2.;
    GLfloat current[4] = { -.5, -halfHeight, 0., 1. };
    if( action->isOfType( SoGLRenderAction::getClassTypeId() ) )
       {
	   glPushAttrib( GL_ALL_ATTRIB_BITS );
	   glDisable( GL_LIGHTING );  // so we don't have to set normals
	   glDepthMask( GL_FALSE );
	   glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
	   glPixelStorei( GL_UNPACK_ROW_LENGTH, icv->width );
	   glRasterPos4fv( current );
	   glPixelZoom( 1, -1 );
	   glDrawPixels( icv -> width,
			 icv -> height, 
			 icv -> backgroundFormat,
			 GL_UNSIGNED_BYTE,
			 icv -> backgroundData );
	   glDepthMask( GL_TRUE );
	   glEnable( GL_LIGHTING );   
	   glPopAttrib();
       }
}


#if 0
void
IvCameraView::interiorOrientationCB( void* icvptr, SoDragger* t1dptr )
{
    IvCameraView* icv = (IvCameraView*) icvptr;
    SoTranslate1Dragger* t1d = (SoTranslate1Dragger*) t1dptr;
    float f = t1d->translation.getValue()[0] / 10. + 1.;
    if( f < .001 )
       f = .001;
    icv -> setInteriorOrientation( 1./f );
}

void
IvCameraView::relativeOrientationCB( void* icvptr, SoDragger* jdptr )
{
    IvCameraView* icv = (IvCameraView*) icvptr;
    SoJackDragger* jd = (SoJackDragger*) jdptr;
    SbMatrix tmpm;
    tmpm.setTransform( jd->translation.getValue(),
		       jd->rotation.getValue(),
		       SbVec3f( 1., 1., 1. ) );
    icv -> setReferenceTransform( tmpm );
}
#endif

void
IvCameraView::interiorOrientationSensorCB( void* icvptr, SoSensor* sensor )
{
    IvCameraView* icv = (IvCameraView*) icvptr;
    SoNodeSensor* nodesensor = (SoNodeSensor*) sensor;
    SoPerspectiveCamera* pc = 
       (SoPerspectiveCamera*) nodesensor -> getAttachedNode();
    float beta = tan( pc->heightAngle.getValue() / 2. ) / .375;
    icv -> setInteriorOrientation( beta );
}

void
IvCameraView::referenceTransformSensorCB( void* icvptr, SoSensor* sensor )
{
    IvCameraView* icv = (IvCameraView*) icvptr;
    SoNodeSensor* nodesensor = (SoNodeSensor*) sensor;
    SoTransform* rt = 
       (SoTransform*) nodesensor -> getAttachedNode();
    SoGetMatrixAction *fuckingstupid = new SoGetMatrixAction( SbViewportRegion() );
    fuckingstupid->apply( rt );
    icv -> setCoordinateTransform( fuckingstupid->getInverse() );
    delete fuckingstupid;
}

void
IvCameraView::coordinateTransformSensorCB( void* icvptr, SoSensor* sensor )
{
    IvCameraView* icv = (IvCameraView*) icvptr;
    SoNodeSensor* nodesensor = (SoNodeSensor*) sensor;
    SoTransform* ct = 
       (SoTransform*) nodesensor -> getAttachedNode();
    SoGetMatrixAction *fuckingstupid = new SoGetMatrixAction( SbViewportRegion() );
    fuckingstupid->apply( ct );
    icv -> setCoordinateTransform( fuckingstupid->getMatrix() );
    delete fuckingstupid;
}

void
IvCameraView::keypressCB( void* cvptr, SoEventCallback* node )
{
    IvCameraView* cv = (IvCameraView*) cvptr;
    if( cv->userKeyCB )
       if( cv->userKeyCB( cv->userKeyData, node ) )
	  return;
    const SoEvent *event = node -> getEvent();
    if( SO_KEY_PRESS_EVENT( event, Q ) )
       exit(0);
}

void
IvCameraView::mousepressCB( void* cvptr, SoEventCallback* node )
{
    IvCameraView* cv = (IvCameraView*) cvptr;
    if( cv->userMouseCB )
       if( cv->userMouseCB( cv->userMouseData, node ) )
	  return;
#if 0
    const SoEvent *event = node -> getEvent();
    if( SO_MOUSE_PRESS_EVENT( event, BUTTON1 ) )
       cerr << "IvCameraView: Button 1" << endl;
#endif
}

void
IvCameraView::handle()
{
    XtInputMask m;
    while( m = XtAppPending( SoXt::getAppContext() ) )
       XtAppProcessEvent( SoXt::getAppContext(), m );
}

void
IvCameraView::setBackgroundColor( float r, float g, float b )
{
    renderArea -> setBackgroundColor( SbColor( r, g, b ) );
}

void
IvCameraView::setLightModel( int model )
{
    lightModel -> model.setValue( model );
}

void
IvCameraView::setSize( int h, int w, int y, int x )
{
    XtConfigureWidget( getWidget(), x, y, w, h, 0 );
    height = h;
    width = w;
    float b = -1. / ( camera -> position.getValue()[2] );
    float halfHeight = float(height) / float(width) / 2.;
    camera -> heightAngle = 2. * atan( halfHeight * b );
    camera -> aspectRatio = float(width) / float(height);
}
