// qmLUDecomposition.c++

#include "qmLUDecomposition.h++"

static inline real absVal( real x ) { return( ( x < 0. ) ? -x : x ); }
const real SMALLFLOAT = 1.0E-14;

int
qmLUDecomposition::findMaxPivot( qmMatrixRef& A, int opRow, int *rowOrder )
{
    // We are trying to find the "opRow"th row, having already processed
    // rows 0 through opRow-1
    // This function finds the row with the largest value in column "opRow"
    // and puts that row next in the rowOrder table
    // This is an inner loop of decompose and does not really make
    // sense on its own

    int height = A.getHeight();
    int maxRow = opRow;
    real maxValue = 0.;
    for ( int i = opRow; i < height; i++) 
       {
	   real cmpval = absVal( A( rowOrder[i], opRow ) );
	   if ( cmpval > maxValue ) 
	      {
		  maxRow = i;
		  maxValue = cmpval;
	      }
       }
    if ( maxRow != opRow ) 
       {
	   int tmp_i = rowOrder[ opRow ];
	   rowOrder[ opRow ] = rowOrder[ maxRow ];
	   rowOrder[ maxRow ] = tmp_i;
       }
    if ( maxValue < SMALLFLOAT)
       return 1;
    return 0;
}

int
qmLUDecomposition::LUDecompose( qmMatrixRef& A, int *rowOrder )
{
    int height = A.getHeight();
    
    /* for each row, zero all elements below that diagonal point */

    for (int i = 0; i < height - 1; i++) 
       {
	   if( findMaxPivot( A, i, rowOrder ) ) 
	      {
#ifdef QmWarnings
		  cerr << "Singular qmLUDecomposition?? (aborted)" << endl;
#endif
		  return 1; // Singular??
	      }
	   real *pivotSpot = A.getArray( rowOrder[i], i );
	   real inversePivot = 1. / *pivotSpot;
	   for ( int row = i+1; row < height; row++ ) 
	      {
		  real* multiplierSpot = A.getArray( rowOrder[row], i );
		  *multiplierSpot *= inversePivot;
		  real multiplier = *multiplierSpot++; // watch it! (++) 
		  real* pivotRow = pivotSpot + 1;
		  for( int col = i + 1; col < height; col++ )
		     *multiplierSpot++ -= *pivotRow++ * multiplier;
	      }
       }
    return 0;
}

void
qmLUDecomposition::forwardSubstitute( qmMatrixRef& LU,
				      int *rowOrder,
				      real *b,
				      real *y )

{
    // computes y for Ly = b
    int width = LU.getWidth();
    for ( int i = 0; i < width; i++ ) 
       {
	   int row = rowOrder[i];
	   real* rowptr = LU.getArray( row, 0 );
	   y[i] = b[row];
	   for ( int col = 0; col < i; col++ )
	      y[i] -= *rowptr++ * y[col];
       }
}

void
qmLUDecomposition::backwardSubstitute( qmMatrixRef& LU,
				       int* rowOrder,
				       real *y,
				       real *x )
{
    // computes x for Ux = y
    int height = LU.getWidth();
    for ( int i = height-1; i >= 0; i-- ) // each row
       {
	   int row = rowOrder[i];
	   real* rowptr = LU.getArray( row, i );
	   real pivot = 1. / *rowptr++;
	   real sum = 0.;
	   for ( int col = i + 1; col < height; col++)
	      sum += *rowptr++ * x[col];
	   x[i] = (y[i] - sum) * pivot;
       }
}

// qmLUDecomposition&
int
qmLUDecomposition::set( const qmMatrixRef& m )
{
    LUmatrix.set( m );
    if( rowOrder )
       delete [] rowOrder;
    rowOrder = new int[ m.getWidth() ];
    for( int i = 0; i < m.getWidth(); i++ )
       rowOrder[i] = i;
    return LUDecompose( LUmatrix, rowOrder );
    // return *this;
}

int
qmLUDecomposition::LUSolve( qmMatrixRef& LU, 
			    int* rorder,
			    qmVectorRef& y,
			    qmVectorRef& x )
{
    qmVector c( y.getLength() );
    forwardSubstitute( LU, rorder, y, c );
    backwardSubstitute( LU, rorder, c, x );
    return 0;
}

int
qmLUDecomposition::LUInvert( qmMatrixRef& LU, int* rorder, qmMatrixRef& Ai )
{
    // use the A=LU decomp to compute Ai (inverse of A)
    int width = LU.getWidth();
    qmVector y( width );
    qmVector c( width );
    qmVector x( width );
    y.setZero();
    for( int i = 0; i < width; i++ )
       {
	   y(i) = real(1.);
	   forwardSubstitute( LU, rorder, y, c );
	   backwardSubstitute( LU, rorder, c, x );
	   y(i) = real(0.);
	   for( int j = 0; j < width; j++ )
	      Ai( j, i ) = x( j );
       }
    return 0;
}

int
qmLUDecomposition::solve( qmVectorRef& y, qmVectorRef& x )
{
    return LUSolve( LUmatrix, rowOrder, y, x );
}

int
qmLUDecomposition::invert( qmMatrixRef& Ai )
{
    return LUInvert( LUmatrix, rowOrder, Ai );
}

ostream&
qmLUDecomposition::print( ostream& os )
{
    os << "LU= " << endl << LUmatrix << endl;
    os << "row order=";
    for( int i = 0; i < LUmatrix.getWidth(); i++ )
       os << " " << rowOrder[i];
    os << endl;
    return os;
}

ostream&
operator<<( ostream& os, qmLUDecomposition& l )
{
    return l.print( os );
}
