// qmMatrixRef.c++

#include <math.h>
#include "QuickMath/qmMatrixRef.h++"
#include "QuickMath/qmMatrix.h++"
#include "QuickMath/qmCompileOptions.h++"
#include "QuickMath/qmMinMax.h++"
#include "QuickMath/qmLUDecomposition.h++"
#include "QuickMath/qmEigenStructure.h++"

qmMatrixRef&
qmMatrixRef::set( int h, int w, real* a )
{
#ifdef QmDoBoundaryChecking
    QmAssert( h > 0 && w > 0 && a );
#endif
    elementCount = h*w;
    width = w;
    array = a;
    return *this;
}

qmMatrixRef&
qmMatrixRef::set( const qmMatrixRef& mr )
{
    memcpy( this, &mr, sizeof( qmMatrixRef ) );
    return *this;
}

qmMatrixRef&
qmMatrixRef::copy( const qmMatrixRef& mr )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount && 
	      elementCount == mr.elementCount &&
	      width == mr.width );
#endif
    memcpy( array, mr.array, elementCount * sizeof(real) );
    return *this;
}

qmMatrixRef&
qmMatrixRef::operator=( const qmMatrixRef& mr )
{
#ifdef QmWarnings
    cerr << "qmMatrixRef: WARNING: operator=() does shadow, not copy" << endl;
    cerr << "                      use set or copy instead" << endl;
#endif
    return set( mr );
}

void
qmMatrixRef::setData( real *p )
{
#ifdef QmDoBoundaryChecking
    QmAssert( p != (real *) NULL);
    QmAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register real *to = array;
    while( count-- )
       (*to++) = *p++;
}

void
qmMatrixRef::getData( real *p ) const
{
#ifdef QmDoBoundaryChecking
    QmAssert( p != (real *) NULL);
    QmAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register real *fm = array;
    register real *to = p;
    while( count-- )
       (*to++) = *fm++;
}

qmMatrixRef&
qmMatrixRef::setAdd( const qmMatrixRef& a, const qmMatrixRef& b )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 &&
	    elementCount == a.elementCount &&
	    width == a.width &&
	    elementCount == b.elementCount &&
	    width == b.width );
#endif
    real *to = array;
    real *fma = a.array;
    real *fmb = b.array;
    int count = elementCount;
    while( count-- )
       (*to++) = (*fma++) +  (*fmb++);
    return *this;
}

qmMatrixRef&
qmMatrixRef::setSubtract( const qmMatrixRef& a, const qmMatrixRef& b )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 &&
	    elementCount == a.elementCount &&
	    width == a.width &&
	    elementCount == b.elementCount &&
	    width == b.width );
#endif
    real *to = array;
    real *fma = a.array;
    real *fmb = b.array;
    int count = elementCount;
    while( count-- )
       (*to++) = (*fma++) -  (*fmb++);
    return *this;
}

qmMatrixRef&
qmMatrixRef::setMultiply( const qmMatrixRef& a, const qmMatrixRef& b )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 &&
	    width == b.width &&
	    a.width * b.width == b.elementCount &&
	    elementCount * a.width == a.elementCount * width );
#endif
    register real *to = array;
    register real *fma = a.array;
    register real *fmb = b.array;
    int height = elementCount / width;
    for( int i = 0; i < height; i++ )
       for( int j = 0; j < width; j++ )
	  {
	      real sum = 0.;
	      fma = a._getArray( i, 0 );
	      fmb = b._getArray( 0, j );
	      for( int k = 0; k < a.width; k++ )
		 sum += (*fma++) * (*fmb), fmb += b.width;
	      *to++ = sum;
	  }
    return *this;
}

qmMatrixRef&
qmMatrixRef::setMultiplyT( const qmMatrixRef& a, const qmMatrixRef& b )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 &&
	    width > 0 && a.width > 0 && b.width > 0 &&
	    getHeight() == a.getHeight() &&
	    width == b.getHeight() &&
	    a.width == b.width );
#endif
    register real *to = array;
    register real *fma = a.array;
    register real *fmb = b.array;
    int height = elementCount / width;
    for( int i = 0; i < height; i++ )
       for( int j = 0; j < width; j++ )
	  {
	      real sum = 0.;
	      fma = a._getArray( i, 0 );
	      fmb = b._getArray( j, 0 );
	      for( int k = 0; k < a.width; k++ )
		 sum += (*fma++) * (*fmb++);
	      *to++ = sum;
	  }
    return *this;
}

qmMatrixRef&
qmMatrixRef::setTMultiply( const qmMatrixRef& a, const qmMatrixRef& b )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 &&
	    width > 0 && a.width > 0 && b.width > 0 &&
	    getHeight() == a.width &&
	    width == b.width &&
	    a.getHeight() == b.getHeight() );
#endif
    register real *to = array;
    register real *fma = a.array;
    register real *fmb = b.array;
    int height = elementCount / width;
    int inner = a.getHeight();
    for( int i = 0; i < height; i++ )
       for( int j = 0; j < width; j++ )
	  {
	      real sum = 0.;
	      fma = a._getArray( 0, i );
	      fmb = b._getArray( 0, j );
	      for( int k = 0; k < inner; k++ )
		 sum += (*fma) * (*fmb), fma+=a.width, fmb+=b.width;
	      *to++ = sum;
	  }
    return *this;
}

qmMatrixRef&
qmMatrixRef::setDivide( const qmMatrixRef& a, const qmMatrixRef& b )
{
#ifdef QmDoBoundaryChecking
#endif
    qmMatrix c;
    c.setInvert( b );
    return setMultiply( a, c );
}

qmMatrixRef&
qmMatrixRef::setNegate( const qmMatrixRef& a )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 &&
	      elementCount == a.elementCount &&
	      width == a.width );
#endif
    register int count = elementCount;
    register real *to = array;
    register real *fm = a.array;
    while( count-- )
       (*to++) = -(*fm++);
    return *this;
}

qmMatrixRef&
qmMatrixRef::setTranspose( const qmMatrixRef& a )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 &&
	      elementCount == a.elementCount &&
	      width * a.width == elementCount );
#endif
    register real *to = array;
    register real *fm = a.array;
    int row = 0;
    int col = 0;
    for( int el = 0; el < elementCount; el++ )
       {
	   *to++ = *fm, fm += a.width;
	   if( ++col == width )
	      col = 0, fm = a.array + (++row);
       }
    return *this;
}

qmMatrixRef&
qmMatrixRef::setShift( real val, const qmMatrixRef& a )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 &&
	    elementCount == a.elementCount &&
	    width == a.width );
#endif
    register int count = elementCount;
    register real *to = array;
    register real *fm = a.array;
    while( count-- )
       (*to++) = val + (*fm++);
    return *this;
}

qmMatrixRef&
qmMatrixRef::setScale( real val, const qmMatrixRef& a )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 &&
	    elementCount == a.elementCount &&
	    width == a.width );
#endif
    register int count = elementCount;
    register real *to = array;
    register real *fm = a.array;
    while( count-- )
       (*to++) = val * (*fm++);
    return *this;
}

qmMatrixRef&
qmMatrixRef::setScaleAddEq( real val, const qmMatrixRef& a )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 &&
	      elementCount == a.elementCount &&
	      width == a.width );
#endif
    register int count = elementCount;
    register real *to = array;
    register real *fm = a.array;
    while( count-- )
       (*to++) += val * (*fm++);
    return *this;
}

qmMatrixRef&
qmMatrixRef::setAddEq( const qmMatrixRef& a )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 &&
	      elementCount == a.elementCount &&
	      width == a.width );
#endif
    real *to = array;
    real *fm = a.array;
    int count = elementCount;
    while( count-- )
       (*to++) += (*fm++);
    return *this;
}

qmMatrixRef&
qmMatrixRef::setSubtractEq( const qmMatrixRef& a )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 &&
	      elementCount == a.elementCount &&
	      width == a.width );
#endif
    real *to = array;
    real *fm = a.array;
    int count = elementCount;
    while( count-- )
       (*to++) -= (*fm++);
    return *this;
}

qmMatrixRef&
qmMatrixRef::setNegateEq()
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 );
#endif
    real *to = array;
    real *fm = array;
    int count = elementCount;
    while( count-- )
       (*to++) = -(*fm++);
    return *this;
}

qmMatrixRef&
qmMatrixRef::setInvert( const qmMatrixRef& from )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 &&
	    elementCount == from.elementCount &&
	    width == from.width &&
	    width*width == elementCount );
#else
    QmAssert( getWidth() == from.getWidth() && 
	      getHeight() == from.getHeight() &&
	      getWidth() == getHeight() );
#endif
    // qmLUD lud( from );
    // lud.invert( *this );
    qmLUD lud;
    if( lud.set( from ) )
       cerr << "qmMatrixRef::setInvert: singular inversion (aborted)" << endl;
    else
       lud.invert( *this );
    return *this;
}

qmMatrixRef&
qmMatrixRef::setInvertEq()
{
#ifdef QmDoBoundaryChecking
    // check done in setInvert()
#endif
    return setInvert( *this );
}

int
qmMatrixRef::setInvertCheck( const qmMatrixRef& from )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 &&
	    elementCount == from.elementCount &&
	    width == from.width &&
	    width*width == elementCount );
#else
    QmAssert( getWidth() == from.getWidth() && 
	      getHeight() == from.getHeight() &&
	      getWidth() == getHeight() );
#endif
    qmLUD lud;
    if( lud.set( from ) )
       return 1;
    lud.invert( *this );
    return 0;
}

int
qmMatrixRef::setInvertEqCheck()
{
#ifdef QmDoBoundaryChecking
    // check done in setInvert()
#endif
    return setInvertCheck( *this );
}

qmMatrixRef&
qmMatrixRef::setScaleEq( real val )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register real *rp = array;
    while( count-- )
       (*rp++) *= val;
    return *this;
}

qmMatrixRef&
qmMatrixRef::setConst( real val )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register real *rp = array;
    while( count-- )
       (*rp++) = val;
    return *this;
}

qmMatrixRef&
qmMatrixRef::setDiagonal( real val )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 );
#endif
    int height = elementCount / width;
    int count = ( width < height ) ? width : height;
    setZero();
    while( count-- )
       (*this)( count, count ) = val;
    return *this;
}

qmMatrixRef&
qmMatrixRef::setDiagonal( const qmVectorRef& vr )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 && width == vr.elementCount );
#endif
    setZero();
    for( int i = 0; i < width; i++ )
       _el( i, i ) = vr._el( i );
    return *this;
}

void
qmMatrixRef::getDiagonal( const qmVectorRef& vr )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 && width == vr.elementCount );
#endif
    for( int i = 0; i < width; i++ )
       vr._el( i ) = _el( i, i );
}

qmMatrixRef&
qmMatrixRef::setOuterProduct( const qmVectorRef& v1, const qmVectorRef& v2 )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 &&
	      width == v2.getLength() &&
	      width * v1.getLength() == elementCount );
#endif
    int count = elementCount;
    real* to = array;
    real* fmrow = v1._getArray();
    real* fmcol = v2._getArray();
    int col = 0;
    while( count-- )
       {
	   *to++ = *fmrow * (*fmcol++);
	   if( ++col == width )
	      col = 0, fmrow++, fmcol = v2._getArray();
       }
    return *this;
}

qmMatrixRef&
qmMatrixRef::setAddScaledOuterProductEq( const qmVectorRef& v1,
					 const qmVectorRef& v2,
					 real w )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 &&
	      width == v2.getLength() &&
	      width * v1.getLength() == elementCount );
#endif
    int count = elementCount;
    real* to = array;
    real* fmrow = v1._getArray();
    real* fmcol = v2._getArray();
    int col = 0;
    while( count-- )
       {
	   *to++ += w * (*fmrow) * (*fmcol++);
	   if( ++col == width )
	      col = 0, fmrow++, fmcol = v2._getArray();
       }
    return *this;
}

qmMatrixRef&
qmMatrixRef::setVectorScale( const qmVectorRef& v, const qmMatrixRef& m )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 &&
	      getHeight() == v.getLength() &&
	      elementCount == m.elementCount &&
	      width == m.width );
#endif
    register real *to = array;
    register real *fmV = v._getArray();
    register real *fmM = m.array;
    register real val;
    int height = getHeight();
    for( int i = 0; i < height; i++ )
       {
	   val = *fmV++;
	   for( int j = 0; j < width; j++ )
	      (*to++) = val * (*fmM++);
       }
    return *this;
}

real
qmMatrixRef::getDeterminant3() const
{
#ifdef QmDoBoundaryChecking
#endif
    QmAssert( width == 3 && elementCount == 9 );
    real d = 0.;
    d += array[0] * (array[4]*array[8] - array[5]*array[7] );
    d += array[1] * (array[5]*array[6] - array[3]*array[8] );
    d += array[2] * (array[3]*array[7] - array[4]*array[6] );
    return d;
}

real
qmMatrixRef::getDeterminant2() const
{
#ifdef QmDoBoundaryChecking
#endif
    QmAssert( width == 2 && elementCount == 4 );
    return( array[0]*array[3] - array[1]*array[2] );
}

real
qmMatrixRef::getDeterminant() const
{
#ifdef QmDoBoundaryChecking
#endif
    switch( width )
       {
	  case 1:
	   return array[0];
	  case 2:
	   return getDeterminant2();
	  case 3:
	   return getDeterminant3();
	  default:
	    QmAssert( elementCount == width*width );
	    qmEigenStructure eigen( *this );
	    real det = 1.;
	    for (int i=0; i<width; i++)
	      det *= eigen.getEigValues()(i);
	    return (det);
       }
}

ostream&
qmMatrixRef::print( ostream& os, char *lineend ) const
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register real *rp = array;
#if 0
    os << "e= " << elementCount << ", w= " << width 
       << ", a= " << array << endl;
#endif
    for( ; count; count -= width )
       {
	   for( int i = 0; i < width; i++ )
	      os << " " << *rp++;
	   os << ((lineend)?lineend:"\0");
       }
    return os;
}

istream&
qmMatrixRef::scan( istream& is )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register real *rp = array;
    while( count-- )
       is >> *rp++;
    return is;
}

ostream&
qmMatrixRef::write( ostream& os ) const
{
    return( os.write( (char *)array, elementCount * sizeof(real) ) );
}

istream&
qmMatrixRef::read( istream& is )
{
    return( is.read( (char *)array, elementCount * sizeof(real) ) );
}

ostream&
operator<<( ostream& os, const qmMatrixRef& qm )
{
    return qm.print( os );
}

istream&
operator>>( istream& is, qmMatrixRef& qm )
{
    return qm.scan( is );
}

int
qmMatrixRef::operator==( const qmMatrixRef& op ) const
{
    if( elementCount != op.elementCount || width != op.width )
       return 0;
    for( int i = 0; i < elementCount; i++ )
       if( op.array[i] != array[i] )
	  return 0;
    return 1;
}

const qmVectorRef&
qmMatrixRef::getColumn( int i ) const
{
    static qmVector column;
    int height = getHeight();
    column.set( height );
    for( int j = 0; j < height; j++ )
       column( j ) = array[ j*width + i ];
    return column;
}

const qmVectorRef&
qmMatrixRef::getRow( int i ) const
{
    static qmVectorRef row;
    row.set( width, array+i*width );
    return row;
}

void
qmMatrixRef::setColumn( int i, const qmVectorRef& column )
{
    int height = getHeight();
    for( int j = 0; j < height; j++ )
       array[ j*width + i ] = column.getValue( j );
}

void
qmMatrixRef::setRow( int i, const qmVectorRef& row )
{
    qmVectorRef( width, array+i*width ).copy( row );
}
