// qmSingularValueDecomposition.c++

#include "QuickMath/qmSingularValueDecomposition.h++"
#include "qmMinMax.h++"

qmSingularValueDecomposition&
qmSingularValueDecomposition::set( qmMatrixRef& m )
{
    Umatrix.set( m.getHeight(), m.getWidth() );
    singularValues.set( m.getWidth() );
    Vmatrix.set( m.getWidth(), m.getWidth() );
    svd( m, Umatrix, singularValues.getArray(), Vmatrix );
    return *this;
}


static inline real square( real a ) { return a*a; }
static inline real sign( real a, real b )
{
    return( (b >= 0.) ? fabs(a) : -fabs(a) );
}

real
qmSingularValueDecomposition::safePythag( const real a, const real b )
{
    real absa = fabs(a);
    real absb = fabs(b);
    if ( absa > absb )
       return( absa*sqrt( 1. + square(absb/absa ) ) );
    else if ( absb > 0. )
       return( absb*sqrt( 1. + square(absa/absb ) ) );
    return 0.;
}

real
qmSingularValueDecomposition::svdMakeBidiagonal( qmMatrixRef& A,
						 real rowvec[],
						 real w[] )
// A is same size as this and will be altered
// rowvec is like one row of this, i.e. (1,width)
// w is the vector of singular values
// returns norm of A
{
    int height = A.getHeight();
    int width = A.getWidth();
    real f;
    real g = 0.;
    real h;
    real scale = 0.;
    real aNorm = 0.;
    int nextCol;
    for( int col = 0; col < width; col++ )
       {
	   nextCol = col + 1;
	   rowvec[col] = scale * g;
	   g = 0.;
	   // s = 0.;
	   scale = 0.;
	   if ( col < height )
	      {
		  for ( int k = col; k < height; k++ )
		     scale += fabs( A(k,col) );
		  if ( scale ) 
		     {
			 real sumSquares = 0.;
			 for ( k = col; k < height; k++ ) 
			    {
				A(k,col) /= scale;
				sumSquares += A(k,col)*A(k,col);
			    }
			 f = A(col,col);
			 g = -sign( sqrt( sumSquares ), f );
			 h = f * g - sumSquares;
			 A(col,col) = f - g;
			 for ( int j = nextCol; j < width; j++ ) 
			    {
				real s = 0.;
				for ( k = col; k < height; k++)
				   s += A(k,col) * A(k,j);
				f = s / h;
				for ( k = col; k < height; k++ )
				   A(k,j) += f * A(k,col);
			    }
			 for ( k = col; k < height; k++) 
			    A(k,col) *= scale;
		     }
	      }
	   w[col] = scale * g;
	   g = 0.;
	   scale = 0.;
	   if ( col < height && col != width-1  ) 
	      {
		  for ( int k = nextCol; k < width; k++ )
		     scale += fabs(A(col,k));
		  if ( scale ) 
		     {
			 real sumSquares = 0.;
			 for ( k = nextCol; k < width; k++ )
			    {
				A(col,k) /= scale;
				sumSquares += A(col,k)*A(col,k);
			    }
			 f = A(col,nextCol);
			 g = -sign( sqrt( sumSquares ), f );
			 h = f * g - sumSquares;
			 A(col,nextCol) = f - g;
			 for ( k = nextCol; k < width; k++ )
			    rowvec[k] = A(col,k) / h;
			 for ( int j = nextCol; j < height; j++ ) 
			    {
				real s = 0.;
				for ( k = nextCol; k < width; k++ )
				   s += A(j,k) * A(col,k);
				for ( k = nextCol; k < width; k++ )
				   A(j,k) += s * rowvec[k];
			    }
			 for ( k = nextCol; k < width; k++ )
			    A(col,k) *= scale;
		     }
	      }
	   aNorm = max( (double) aNorm,
			(double) ( fabs( w[col] ) + fabs( rowvec[col] ) ) );
       }
    return aNorm;
}

qmMatrixRef&
qmSingularValueDecomposition::svdMakeRHXforms( qmMatrixRef& A,
					       real rowvec[],
					       qmMatrixRef& V )
// rowvec?
// A is the bidiagonal Househoulder-reduced matrix
// V is the output right-hand transformation.
{
    int width = A.getWidth();
    real g = 0.;
    int nextCol = 0;
    for ( int col = width-1; col >= 0; col--) 
       {
	   if ( col < width-1 )
	      {
		  if ( g )
		     {
			 for ( int j = nextCol; j < width; j++ )
			    V( j, col ) = ( A(col,j) / A(col,nextCol) ) / g;
			 
			 for ( j = nextCol; j < width; j++ ) 
			    {
				real s = 0.;
				for ( int k = nextCol; k < width; k++ )
				   s += A(col,k) * V(k,j);
				for ( k = nextCol; k < width; k++ )
				   V( k, j ) += s * V(k,col);
			    }
		     }
		  for ( int j = nextCol; j < width; j++ )
		     V( col, j ) = V( j, col ) = 0.;
	      }
	   V( col, col ) = 1.;
	   g = rowvec[col];
	   nextCol = col;
       }
    return V;
}

qmMatrixRef&
qmSingularValueDecomposition::svdMakeLHXforms( qmMatrixRef& U, real w[] )
{
    int height = U.getHeight();
    int width = U.getWidth();
    for ( int col = min(width,height)-1; col >= 0; col-- ) 
       {
	   int nextCol = col+1;
	   real g = w[col];
	   for ( int j = nextCol; j < width; j++ )
	      U(col,j) = 0.;
	   if (g) 
	      {
		  g = 1. / g;
		  for ( j = nextCol; j < width; j++ ) 
		     {
			 real s = 0.;
			 for ( int k = nextCol; k < height; k++ )
			    s += U(k,col) * U(k,j);
			 real f = ( s / U(col,col) ) * g;
			 for ( k = col; k < height; k++ )
			    U(k,j) += f * U(k,col);
		     }
		  for ( j = col; j < height; j++)
		     U(j,col) *= g;
	      }
	   else
	      for ( j = col; j < height; j++)
		 U(j,col) = 0.;
	   ++U(col,col);
       }
    return U;
}

void
qmSingularValueDecomposition::svdMakeDiagonal( qmMatrixRef& U,
					       real w[],
					       qmMatrixRef& V,
					       real rowvec[],
					       real aNorm )
{
    int width = U.getWidth();
    int height = U.getHeight();
    for( int col = width-1; col >= 0; col-- ) 
       {
	   for ( int iterCount = 0; iterCount < 30; iterCount++ ) 
	      {
		  int nm = 0;
		  int flag = 1;
		  for( int prevCol = col; prevCol >= 0; prevCol-- ) 
		     {
			 nm = prevCol - 1;
			 // note rowvec[0] = 0. always
			 if( fabs( rowvec[prevCol]) + aNorm == aNorm )
			    {
				flag = 0;
				break;
			    }
			 if( fabs( w[nm] ) + aNorm == aNorm ) 
			    break;
		     }
		  // carry over "prevCol"
		  if( flag ) 
		     {
			 real c = 0.;
			 real s = 1.;
			 for ( int i = prevCol; i <= col; i++ ) 
			    {
				real f = s * rowvec[i];
				rowvec[i] = c * rowvec[i];
				if ( (fabs(f) + aNorm) == aNorm) 
				   break;
				real g = w[i];
				real h = safePythag( f, g );
				w[i] = h;
				h = 1. / h;
				c = g * h;
				s = -f * h;
				for( int j = 0; j < height; j++ ) 
				   {
				       real y = U( j, nm );
				       real z = U( j, i );
				       U( j, nm ) = y*c + z*s;
				       U( j, i ) = z*c - y*s;
				   }
			    }
		     }
		  real z = w[col];
		  if (prevCol == col) 
		     {
			 if( z < 0. ) 
			    {
				w[col] = -z;
				for( int j = 0; j < width; j++ ) 
				   V( j, col ) = -V( j, col );
			    }
			 break; // converged, stop iterating
		     }
		  QmAssert( iterCount < 30 );
		  real x = w[prevCol];
		  nm = col - 1;
		  real y = w[nm];
		  real g = rowvec[nm];
		  real h = rowvec[col];
		  real f = ((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y);
		  g = safePythag( f, 1. );
		  f = ((x-z)*(x+z)+h*((y/(f+sign(g,f)))-h))/x;
		  real c = 1.;
		  real s = 1.;
		  for( int j = prevCol; j <= nm; j++ ) 
		     {
			 int i = j+1;
			 g = rowvec[i];
			 y = w[i];
			 h = s*g;
			 g = c*g;
			 real z = safePythag( f, h );
			 rowvec[j] = z;
			 c = f/z;
			 s = h/z;
			 f = x*c + g*s;
			 g = g*c - x*s;
			 h = y*s;
			 y *= c;
			 for( int jj = 0; jj < width; jj++ ) 
			    {
				x = V( jj, j );
				z = V( jj, i );
				V( jj, j ) = x*c + z*s;
				V( jj, i ) = z*c - x*s;
			    }
			 z = safePythag( f, h );
			 w[j] = z;
			 if (z) 
			    {
				z = 1. / z;
				c = f*z;
				s = h*z;
			    }
			 f = c*g + s*y;
			 x = c*y - s*g;
			 for( jj = 0; jj < height; jj++ ) 
			    {
				y = U( jj, j );
				z = U( jj, i );
				U( jj, j ) = y*c + z*s;
				U( jj, i ) = z*c - y*s;
			    }
		     }
		  rowvec[prevCol] = 0.;
		  rowvec[col] = f;
		  w[col] = x;
	      }
       }
}

real*
qmSingularValueDecomposition::svd( qmMatrixRef& A,
				   qmMatrixRef& U,
				   real w[],
				   qmMatrixRef& V )
{
#ifdef QmDoBoundaryChecking
    QmAssert( U.getElementCount() == A.getElementCount() &&
	    U.getWidth() == A.getWidth() );
#endif
    U.copy( A );
    real* rowvec = new real[ A.getWidth() ];
    real aNorm = svdMakeBidiagonal( U, rowvec, w );
    svdMakeRHXforms( U, rowvec, V );
    svdMakeLHXforms( U, w );
    svdMakeDiagonal( U, w, V, rowvec, aNorm );
    delete [] rowvec;
    return w;
}

ostream&
qmSingularValueDecomposition::print( ostream& os )
{
    os << "U= " << endl << Umatrix << endl;
    os << "w= " << endl << singularValues << endl;
    os << "V= " << endl << Vmatrix << endl;
    return os;
}

ostream&
operator<<( ostream& os, qmSingularValueDecomposition& s )
{
    return s.print( os );
}
