// qmVectorRef.c++

#include "QuickMath/qmVectorRef.h++"
#include "QuickMath/qmCompileOptions.h++"

qmVectorRef&
qmVectorRef::set( int n, real* a )
{
    elementCount = n;
    array = a;
    return *this;
}

qmVectorRef&
qmVectorRef::set( const qmVectorRef& vr )
{
    memcpy( this, &vr, sizeof( qmVectorRef ) );
    return *this;
}

qmVectorRef&
qmVectorRef::copy( const qmVectorRef& vr )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount && elementCount == vr.elementCount );
#endif
    memcpy( array, vr.array, elementCount * sizeof(real) );
    return( *this );
}

qmVectorRef&
qmVectorRef::operator=( const qmVectorRef& vr )
{
#ifdef QmWarnings
    cerr << "qmVectorRef: WARNING: operator=() does shadow, not copy" << endl;
    cerr << "                      use set or copy instead" << endl;
#endif
    return set( vr );
}

void
qmVectorRef::setData( const real *p )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 );
    QmAssert( array != (real *) NULL );
#endif
    register int count = elementCount;
    register real *to = array;
    while( count-- )
       (*to++) = *p++;
}

void
qmVectorRef::getData( real *p ) const
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 );
    QmAssert( array != (real *) NULL) ;
#endif
    register int count = elementCount;
    register real *fm = array;
    while( count-- )
       *p++ = *fm++;
}

qmVectorRef&
qmVectorRef::setAdd( const qmVectorRef& a, const qmVectorRef& b )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 );
    QmAssert( elementCount == a.elementCount );
    QmAssert( a.elementCount == b.elementCount );
#endif
    register int count = elementCount;
    register real *to = array;
    register real *fm1 = a.array;
    register real *fm2 = b.array;
    while( count-- )
       (*to++) = (*fm1++) + (*fm2++);
    return *this;
}

qmVectorRef&
qmVectorRef::setSubtract( const qmVectorRef& a, const qmVectorRef& b )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 );
    QmAssert( elementCount == a.elementCount );
    QmAssert( a.elementCount == b.elementCount );
#endif
    register int count = elementCount;
    register real *to = array;
    register real *fm1 = a.array;
    register real *fm2 = b.array;
    while( count-- )
       (*to++) = (*fm1++) - (*fm2++);
    return *this;  
}

qmVectorRef&
qmVectorRef::setNegate( const qmVectorRef& v )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 &&
	      elementCount == v.elementCount);
#endif
    register int count = elementCount;
    register real *to = array;
    register real *from = v.array;
    while( count-- )
       (*to++) = -(*from++);
    return *this;
}

qmVectorRef&
qmVectorRef::setShift( real val, const qmVectorRef& v )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 &&
	      elementCount == v.elementCount);
#endif
    register int count = elementCount;
    register real *to = array;
    register real *from = v.array;
    while( count-- )
       (*to++) = val + (*from++);
    return *this;
}

qmVectorRef&
qmVectorRef::setScale( real val, const qmVectorRef& v )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 &&
	      elementCount == v.elementCount);
#endif
    register int count = elementCount;
    register real *to = array;
    register real *from = v.array;
    while( count-- )
       (*to++) = val * (*from++);
    return *this;
}

qmVectorRef&
qmVectorRef::setScaleAddEq( real w, const qmVectorRef& v )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 && elementCount == v.elementCount);
#endif
    register int count = elementCount;
    register real *to = array;
    register real *from = v.array;
    while( count-- )
       (*to++) += w * (*from++);
    return *this;
}

qmVectorRef&
qmVectorRef::setAddEq( const qmVectorRef& op )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 );
    QmAssert( elementCount == op.elementCount );
#endif
    register int count = elementCount;
    register real *to = array;
    register real *fm = op.array;
    while( count-- )
       (*to++) += *fm++;
    return *this;
}

qmVectorRef&
qmVectorRef::setSubtractEq( const qmVectorRef& op )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 );
    QmAssert( elementCount == op.elementCount );
#endif
    register int count = elementCount;
    register real *to = array;
    register real *fm = op.array;
    while( count-- )
       (*to++) -= *fm++;
    return *this;
}

qmVectorRef&
qmVectorRef::setNegateEq()
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register real *to = array;
    register real *fm = array;
    while( count-- )
       (*to++) = -(*fm++);
    return *this;
}

qmVectorRef&
qmVectorRef::setShiftEq( real val )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register real *rp = array;
    while( count-- )
       (*rp++) += val;
    return *this;
}

qmVectorRef&
qmVectorRef::setScaleEq( real val )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register real *rp = array;
    while( count-- )
       (*rp++) *= val;
    return *this;
}

qmVectorRef&
qmVectorRef::setConst( real val )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register real *rp = array;
    while( count-- )
       (*rp++) = val;
    return *this;
}

qmVectorRef&
qmVectorRef::setMultiply( const qmMatrixRef& m, const qmVectorRef& v )
{
#ifdef QmDoBoundaryChecking
    QmAssert( v.getLength() == m.getWidth() &&
	      v.getLength() * getLength() == m.getElementCount() );
    // (do not use getHeight(), because you can get divide by zero)
#endif
    register real *to = getArray();
    register real *fmM = m._getArray();
    register real *fmV;
    register int count;
    for( int row = 0; row < getLength(); row++ )
       {
	   *to = 0.;
	   fmV = v._getArray();
	   count = m.getWidth();
	   while( count-- )
	      *to += (*fmM++) * (*fmV++);
	   to++;
       }
    return *this;
}

qmVectorRef&
qmVectorRef::setTMultiply( const qmMatrixRef& m, const qmVectorRef& v )
{
#ifdef QmDoBoundaryChecking
    QmAssert( getLength() == m.getWidth() &&
	      v.getLength() * getLength() == m.getElementCount() );
#endif
    register real *to = getArray();
    register real *fmM;
    register real *fmV;
    register int count;
    for( int col = 0; col < getLength(); col++ )
       {
	   *to = 0.;
	   fmV = v._getArray();
	   fmM = m._getArray( 0, col );
	   count = v.getLength();
	   while( count-- )
	      *to += (*fmM) * (*fmV++), fmM+=m.getWidth();
	   to++;
       }
    return *this;
}

qmVectorRef&
qmVectorRef::setVectorScale( const qmVectorRef& a, const qmVectorRef& b )
{
#ifdef QmDoBoundaryChecking
    QmAssert( getLength() == a.getLength() &&
	      a.getLength() == b.getLength() );
#endif
    register real *to = getArray();
    register real *fmA = a._getArray();
    register real *fmB = b._getArray();
    register int count = elementCount;
    while( count-- )
       (*to++) = (*fmA++) * (*fmB++);
    return *this;
}

real
qmVectorRef::getInnerProduct( const qmVectorRef& op ) const
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 );
    QmAssert( elementCount == op.elementCount );
#endif
    real out = 0.;
    register int count = elementCount;
    register real *fm1 = array;
    register real *fm2 = op.array;
    while( count-- )
       out += (*fm1++) * (*fm2++);
    return out;
}

int
qmVectorRef::operator==( const qmVectorRef& op ) const
{
    if( elementCount != op.elementCount )
       return 0;
    for( int i = 0; i < elementCount; i++ )
       if( op.array[i] != array[i] )
	  return 0;
    return 1;
}

qmVectorRef&
qmVectorRef::setAbsolute( )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register real *rp = array;
    while( count-- )
       {
	   if( *rp < 0 )
	      *rp = (-(*rp));
	   rp++;
       }
    return *this;
}

ostream&
qmVectorRef::print( ostream& os ) const
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register real *rp = array;
    while( count-- )
       os << " " << *rp++;
    return os;
}

istream&
qmVectorRef::scan( istream& is )
{
#ifdef QmDoBoundaryChecking
    QmAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register real *rp = array;
    while( count-- )
       is >> *rp++;
    return is;
}

ostream&
qmVectorRef::write( ostream& os ) const
{
    return os.write( (char *)array, elementCount * sizeof(real) );
}


istream&
qmVectorRef::read( istream& is )
{
    return is.read( (char *)array, elementCount * sizeof(real) );
}

qmVectorRef&
qmVectorRef::setCross( const qmVectorRef& a, const qmVectorRef& b )
{
    assert( 3==elementCount && 3==a.elementCount && 3==b.elementCount );
    array[ 0 ] = a.getValue(1)*b.getValue(2) - a.getValue(2)*b.getValue(1);
    array[ 1 ] = a.getValue(2)*b.getValue(0) - a.getValue(0)*b.getValue(2);
    array[ 2 ] = a.getValue(0)*b.getValue(1) - a.getValue(1)*b.getValue(0);
    return *this;
}


ostream&
operator<<( ostream& os, const qmVectorRef& vr ) { return vr.print( os ); }

istream&
operator>>( istream& is, qmVectorRef& vr ) { return vr.scan( is ); }

