// qmVtiKalmanFilter.c++
// 
// Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
// rights reserved.
// 
// Developed  by Ali Azarbayejani and Alex Pentland,  and Martin  Friedmann
// at the Media Laboratory, MIT, Cambridge, Massachusetts.
// 
// Permission to use, copy,  or modify this software and  its documentation
// for  educational  and  research purposes only and without fee  is hereby
// granted, provided  that this copyright notice and the original authors's
// names appear  on all copies and supporting documentation.  If individual
// files are  separated from  this  distribution directory  structure, this
// copyright notice must be included.  For any other uses of this software,
// in original or  modified form, including but not limited to distribution
// in whole or in  part, specific  prior permission  must be  obtained from
// MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
// basis  of  a commercial  software  or  hardware product  without first
// obtaining appropriate licenses  from MIT.  MIT. makes no representations
// about the suitability of this  software for any purpose.  It is provided
// "as is" without express or implied warranty.
// 

#include "qmVtiKalmanFilter.h++"
#include <stdlib.h>

static char *banner = "ERROR: qmVtiKalmanFilter: ";

static
inline
void
Error( char *s )
{
  cerr << banner << s << endl;;
  exit( 1 );
}

static
inline
void
Error( char *s, qmTime& t1, qmTime& t2 )
{
  cerr << banner << s << endl;
  //t1.print( cerr << "\tmmt time " ) << endl;
  //t2.print( cerr << "\test time " ) << endl;
  exit( 1 );
}

qmVtiKalmanFilter::qmVtiKalmanFilter( int ns, real rv, real qv, double to )
  : qmKalmanFilter( ns, rv, qv, 1. ),
    tP(),
    tE( tP ),
    xBase( n ),
    PBase( n, n ),
    dQ( n, n )
{
  if( to >= 0. )
     tP = tE = to;
}

qmVtiKalmanFilter&
qmVtiKalmanFilter::update( const qmVectorRef& y, qmTime& ty, int N )
// mmt y at time ty, use N steps for prediciton
{
  // Generate a prediction x(ty|ty-dt), P(ty|ty-dt)
  //   where ty = time this mmt was taken
  //         dt = prediction time interval
  // (1) Figure out which estimate to use for the prediction
  qmTime base;
  base = tE;
  if( ty < tE )
    {
      Error( "mmt time before previous estimate", ty, tE );
    }
  if( tP > tE && tP <= ty )
     {
       base = tP;
       xBase = xP;
       PBase = Pp;
     }
  else
     {
       xBase = xE;
       PBase = Pe;
     }
  // (2) Figure out the time interval
  qmTime interval( ty - base );
  interval /= N;
  // (3) Propagate the prediction
  dQ.setScale( real(1./N), Q );
  for( int ti = 0; ti < N; ti++ )
     {
       StateTransitionMatrix( base, interval );
       xP.setMultiply( Phi, xBase );
       Pp.setMultiplyT( tempNxN.setMultiply( Phi, PBase ), Phi ).setAddEq( dQ );
       base += interval;
       tP = base;
       xBase = xP;
       PBase = Pp;
     }

  // Generate an estimate x(ty|ty), P(ty|ty) using y
  // (1) K(ty) = P(ty|ty-dt) H"(t) [ H(t) P(ty|ty-dt) H"(t) + R(t) ]^(-1)
  tempNxP.setMultiplyT( Pp, H );
  tempPxP.setMultiply( H, tempNxP );
  tempPxP.setAddEq( R );
  tempPxP.setInvertEq();
  K.setMultiply( tempNxP, tempPxP );
  // (2) x(ty|ty) = x(ty|ty-dt) + K(ty) [ y - H(t) x(ty|ty-dt) ]
  yP.setMultiply( H, xP );
  dY.setSubtract( y, yP );
  dX.setMultiply( K, dY );
  xE.setAdd( xP, dX );
  // (3) P(ty|ty) = ( I - K(ty) H(t) ) P(ty|ty-dt)
  Pe.setMultiply( tempNxN.setMultiply( K, H ).setSubtract( I, tempNxN ), Pp );
  tE = ty;

  return *this;
}

qmMatrixRef&
qmVtiKalmanFilter::StateTransitionMatrix( qmTime& , qmTime& dt )
{
  static int inited = 0;
  if( n % 3 || n != 3*p )
     Error( "this is not an identity filter; need to define STM" );
  if( !inited )
     Phi.setIdentity();
  real halfDtSquared = .5 * dt * dt;
  int ns = n / 3;
  for( int i = 0; i < ns; i++ )
     {
       Phi( i, i+ns ) = dt;
       Phi( i, i+2*ns ) = halfDtSquared;
       Phi( i+ns, i+2*ns ) = dt;
     }
  return Phi;
}
