// noLevenbergMarquardt.c++

#include "noLevenbergMarquardt.h++"

real noLevenbergMarquardt::defaultU = .01;
real noLevenbergMarquardt::defaultV = 10.;
real noLevenbergMarquardt::defaultTolerance = .001;
real noLevenbergMarquardt::defaultAbsTolerance = .001;

noLevenbergMarquardt::noLevenbergMarquardt()
  : u( defaultU ),
    v( defaultV ),
    dy(),
    Rivec(),
    x(),
    J(),
    dx(),
    new_x(),
    new_dy(),
    new_loss( 0. ),
    reduced( 0 ),
    loss( 0. ),
    JtDy(),
    RiDy(),
    RiJ(),
    magJtDy( 0. ),
    gradientTolerance( defaultTolerance ),
    absoluteTolerance( defaultAbsTolerance ),
    solveDx(),
    solveH(),
    solveJtJ(),
    solveuI(),
    solveJtDy(),
    verbose( 0 ),
    iterationLimit( 1000 ),
    echoExitStats( 0 ),
    estimateStatus( None ),
    solveStatus( 0 )
{}

void
noLevenbergMarquardt::init( noErrorFunction& a )
{
    erf = &a;
    u = defaultU;
    v = defaultV;
    dy.set( a.getNdy() );
    Rivec.set( a.getNdy() );
    Rivec.setConst( 1. );
    x.set( a.getNx() );
    J.set( a.getNdy(), a.getNdx() );
    dx.set( a.getNdx() );
    new_x.set( a.getNx() );
    new_dy.set( a.getNdy() );
    new_loss = 0.;
    reduced = 0;
    loss = 0.;
    JtDy.set( a.getNdx() );
    RiDy.set( a.getNdy() );
    RiJ.set( a.getNdy(), a.getNdx() );
    magJtDy = 0.;
    gradientTolerance = defaultTolerance;
    absoluteTolerance = defaultAbsTolerance;
    solveDx.set( a.getNdx() );
    solveH.set( a.getNdx(), a.getNdx() );
    solveJtJ.set( a.getNdx(), a.getNdx() );
    solveuI.set( a.getNdx(), a.getNdx() );
    solveJtDy.set( a.getNdx() );
    verbose = 0;
}

real
noLevenbergMarquardt::computeLoss( qmVectorRef& dy )
{
    return( sqrt( dy.getInnerProduct( RiDy.setVectorScale( Rivec, dy ) ) ) );
}

const
qmVectorRef&
noLevenbergMarquardt::solve( qmMatrixRef& J, qmVectorRef& dy, real u )
{
    solveJtJ.setTMultiply( J, RiJ.setVectorScale( Rivec, J ) );
    solveuI.setDiagonal( u );
    solveH.setAdd( solveJtJ, solveuI );
    if(verbose>20) solveH.print( cerr << "H= " );
    solveJtDy.setTMultiply( J, RiDy.setVectorScale( Rivec, dy ) );
    if(verbose>18) solveJtDy.print( cerr << "JtDy= " ) << endl;
    if( solveH.setInvertEqCheck() )
       {
	   solveStatus = SolveFail;
	   return solveDx;
       }
    solveDx.setMultiply( solveH, solveJtDy );
    solveStatus = SolveSuccess;
    return solveDx.setNegateEq();
}

const
qmVectorRef&
noLevenbergMarquardt::estimate()
{
    int failure = 0;
    noErrorFunction& a = *erf;
    dy = a( x );
    J = a.jacobian( x );
    loss = computeLoss( dy );
    JtDy.setTMultiply( J, RiDy.setVectorScale( Rivec, dy ) );
    magJtDy = sqrt( JtDy.getInnerProduct( JtDy ) );
    int iterationCount = 0;
    while( (magJtDy > gradientTolerance || loss > absoluteTolerance ) 
	   && !failure )
       {
	   // if(verbose) cout << x << endl;
	   if(verbose>5) cerr << "\titer= " << iterationCount << endl;
	   if(verbose>5) cerr << "\tloss= " << loss << endl;
	   if(verbose>5) cerr << "\tgrad= " << magJtDy << endl;
	   if(verbose>16) dy.print( cerr << "dy= " ) << endl;
	   if(verbose>16) J.print( cerr << "J= " ) << endl;
	   do {
	       dx = solve( J, dy, u );
	       if( SolveFail == solveStatus )
		  {
		      estimateStatus = Singular;
		      failure = 1;
		      continue;
		  }
	          if(verbose>10) dx.print( cerr << "\tdx= " ) << endl;
	       new_x = a.add( x, dx );
	          if(verbose>10) new_x.print( cerr << "\tnew_x= " ) << endl;
	       new_dy = a( new_x );
	       new_loss = computeLoss( new_dy );
	          if(verbose>5) 
		  cerr << "\t\tu= " << u << ",  " 
		       << "new_loss= " << new_loss << endl;
	       reduced = ( new_loss < loss );
	       if( reduced )
		  u /= v;
	       else
		  u *= v;
	       if(verbose>5) cerr << "\t\tnew_u= " << u << endl;
	       if( iterationCount++ > iterationLimit )
		  {
		      estimateStatus = Overflow;
		      failure = 1;
		  }
	   }
	   while( !reduced && !failure );
	   x = new_x;
	   dy = new_dy;
	   loss = new_loss;
	   J = a.jacobian( x );
	   JtDy.setTMultiply( J, RiDy.setVectorScale( Rivec, dy ) );
	   magJtDy = sqrt( JtDy.getInnerProduct( JtDy ) );
	   if(verbose>10) J.print( cerr << "\tJ= " ) << endl;
	   if(verbose>10) dy.print( cerr << "\tdy= " ) << endl;
	   if(verbose>10) cerr << "\tmagJtDy= " << magJtDy << endl;
       }
    if( failure )
       {
	   cerr << "noLevenbergMarquardt: estimate failed because of "
		<< ((Overflow==estimateStatus) ? "OVERFLOW"
		    : (Singular==estimateStatus) ? "SINGULAR" : "Whoops")
		<< ", iter= " << iterationCount << endl;
       }
    else
       estimateStatus = Success;
    if( echoExitStats )
       {
	   cerr << "LM: iter= " << iterationCount << ", grad= " << magJtDy
		<< ", loss= " << loss << endl;
       }
    return x;
}
